@charset 'UTF-8';

* {
	font-family: "Source Sans Pro", sans-serif;
}

body {
	/* white_carbon2.png */
	background: #f6f8f9 url("data:image/png;base64,iVBORw0KGgoAAAANSUhEUgAAAAgAAAAIBAMAAAA2IaO4AAAAFVBMVEX19PXo5ujv7u/o5ej29fb18/X29PaHCvlgAAAAIklEQVQIHWNgEARCBiUgNHQQdAAzGAIgYglKDIIOgglgMQA+hAOJM/f1wAAAAABJRU5ErkJggg==");
}

div#doc-content {
	background: rgb(244, 244, 244);
}

/* glowing stuff */
h1.glow, h2.glow, h3.glow,
h4.glow, h5.glow, h6.glow {
	text-shadow: 0 0 15px rgb(183, 28, 28);
}
div.memitem.glow {
	box-shadow: 0 0 15px rgb(183, 28, 28);
}

/* header */
div#titlearea {
	background: #f5f5f5;
	margin-bottom: 20px;
	padding: 10px 0;
	box-shadow: 0 3px 3px rgba(0, 0, 0, 0.35);

	border-bottom: 0;

	background-image: linear-gradient(
		bottom, rgb(239,239,239) 0%, rgb(255,255,255) 100%);
	background-image: -o-linear-gradient(
		bottom, rgb(239,239,239) 0%, rgb(255,255,255) 100%);
	background-image: -moz-linear-gradient(
		bottom, rgb(239,239,239) 0%, rgb(255,255,255) 100%);
	background-image: -webkit-linear-gradient(
		bottom, rgb(239,239,239) 0%, rgb(255,255,255) 100%);
}

div#projectname {
	font-family: 'Source Sans Pro', sans-serif;
	font-size: 25px;
	font-weight: bold;
}

div#projectbrief {
	font-family: 'Source Sans Pro', sans-serif;
	font-size: 13px;
}

/* the first tab bar is different ... */
div.tabs {
	background: #3b3d3e;
	height: 30px;
	font-size: 14px;
}

div.tabs li {
	background: transparent;
	line-height: 30px;
}

div.tabs li:hover {
	background: rgb(46, 46, 46);
}

div.tabs a {
	color: rgba(255, 255, 255, 0.75);
	background: transparent;

	text-transform: uppercase;
	text-shadow: none;
}

div.tabs a:hover {
	color: rgb(233, 233, 233);
	background: transparent;
}

div.tabs li.current a {
	color: rgb(74, 74, 74);
	background: rgb(250, 250, 250);

	text-shadow: none;
}

/* the other tabs are the same */
div.tabs2,
div.tabs3 {
	background: rgb(250, 250, 250);
	border: 1px solid rgb(196, 196, 196);
	border-top: 0;
	font-size: 13px;
}

div.tabs2 li,
div.tabs3 li {
	background: transparent;
	line-height: 30px;
}

div.tabs2 a,
div.tabs3 a {
	padding: 0 8px;
	color: rgb(183, 28, 28);
	background: transparent;
}

div.tabs2 a:hover,
div.tabs3 a:hover {
	color: rgb(74, 74, 74);
	background: transparent;

	text-shadow: none;
}

div.tabs2 li.current a,
div.tabs3 li.current a {
	color: rgb(74, 74, 74);
	background: transparent;
	text-shadow: none;
}

/* nav path (at the bottom of the page) */
div.navpath {
	background: rgb(233, 233, 233);
	border: 1px solid rgb(196, 196, 196);
	border-width: 1px 0;
	font-size: 13px;
	line-height: 20px;
}

div.navpath ul {
	background: transparent;
	border: 0;
}

/* directories */
div.directory {
	border-color: rgba(0, 0, 0, 0.35);
}
