/*
 * Copyright (c) 2014 Hugh Bailey <obs.jim@gmail.com>
 *
 * Permission to use, copy, modify, and distribute this software for any
 * purpose with or without fee is hereby granted, provided that the above
 * copyright notice and this permission notice appear in all copies.
 *
 * THE SOFTWARE IS PROVIDED "AS IS" AND THE AUTHOR DISCLAIMS ALL WARRANTIES
 * WITH REGARD TO THIS SOFTWARE INCLUDING ALL IMPLIED WARRANTIES OF
 * MERCHANTABILITY AND FITNESS. IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR
 * ANY SPECIAL, DIRECT, INDIRECT, OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES
 * WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN
 * ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF
 * OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.
 */

#pragma once

template<typename T> class CoTaskMemPtr {
	T *ptr;

	inline void Clear() {if (ptr) CoTaskMemFree(ptr);}

public:
	inline CoTaskMemPtr()        : ptr(NULL) {}
	inline CoTaskMemPtr(T *ptr_) : ptr(ptr_) {}
	inline ~CoTaskMemPtr()                   {Clear();}

	inline operator T*() const               {return ptr;}
	inline T *operator->() const             {return ptr;}

	inline CoTaskMemPtr& operator=(T* val)
	{
		Clear();
		ptr = val;
	}

	inline T** operator&()
	{
		Clear();
		ptr = NULL;
		return &ptr;
	}
};
