EXPORTS
    json_delete
    json_true
    json_false
    json_null
    json_string
    json_stringn
    json_string_nocheck
    json_stringn_nocheck
    json_string_value
    json_string_length
    json_string_set
    json_string_setn
    json_string_set_nocheck
    json_string_setn_nocheck
    json_integer
    json_integer_value
    json_integer_set
    json_real
    json_real_value
    json_real_set
    json_number_value
    json_array
    json_array_size
    json_array_get
    json_array_set_new
    json_array_append_new
    json_array_insert_new
    json_array_remove
    json_array_clear
    json_array_extend
    json_object
    json_object_size
    json_object_get
    json_object_set_new
    json_object_set_new_nocheck
    json_object_del
    json_object_clear
    json_object_update
    json_object_update_existing
    json_object_update_missing
    json_object_iter
    json_object_iter_at
    json_object_iter_next
    json_object_iter_key
    json_object_iter_value
    json_object_iter_set_new
    json_object_key_to_iter
    json_dumps
    json_dumpf
    json_dump_file
    json_dump_callback
    json_loads
    json_loadb
    json_loadf
    json_load_file
    json_load_callback
    json_equal
    json_copy
    json_deep_copy
    json_pack
    json_pack_ex
    json_vpack_ex
    json_unpack
    json_unpack_ex
    json_vunpack_ex
    json_set_alloc_funcs

